package ru.dbotthepony.kstarbound.world.entities.behavior

import org.apache.logging.log4j.LogManager
import org.classdump.luna.ByteString
import org.classdump.luna.Table
import org.classdump.luna.exec.CallPausedException
import org.classdump.luna.lib.CoroutineLib
import org.classdump.luna.runtime.Coroutine
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeParameter
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState
import java.util.concurrent.atomic.AtomicLong

class DecoratorNode(val name: String, val parameters: Map<String, NodeParameter>, val child: AbstractBehaviorNode) : AbstractBehaviorNode() {
	private var coroutine: Coroutine? = null
	private val nodeID = BehaviorState.NODE_GARBAGE_INDEX.getAndIncrement()

	override fun run(delta: Double, state: BehaviorState): Status {
		var coroutine = coroutine

		if (coroutine == null) {
			val parameters = state.blackboard.parameters(parameters, this)
			val fn = state.functions[name] ?: throw RuntimeException("How? $name")

			try {
				coroutine = state.lua.call(CoroutineLib.create(), fn)[0] as Coroutine
				val result = state.lua.call(CoroutineLib.resume(), coroutine, parameters, state.blackboard, nodeID)
				val status = result[0] as Boolean

				if (!status) {
					LOGGER.warn("Behavior DecoratorNode '$name' failed: ${result[1]}")
					return Status.FAILURE
				}

				if (result.size == 1) {
					val coroutineStatus = state.lua.call(CoroutineLib.status(), coroutine)[0] as String

					if (coroutineStatus == "dead") {
						// quite unexpected, but whatever
						return Status.SUCCESS
					} else {
						this.coroutine = coroutine
					}
				} else {
					val nodeStatus = result.getOrNull(1) as? Boolean ?: false
					// val nodeExtra = result.getOrNull(3) as? Table
					return if (nodeStatus) Status.SUCCESS else Status.FAILURE
				}
			} catch (err: CallPausedException) {
				LOGGER.error("Behavior DecoratorNode '$name' called blocking code, which initiated pause. This is not supported.")
				return Status.FAILURE
			}
		}

		// decorator runs its child on yield and is resumed with the child's status on success or failure
		var status = Status.RUNNING

		while (status == Status.RUNNING) {
			val childStatus = child.runAndReset(delta, state)

			if (childStatus == Status.SUCCESS || childStatus == Status.FAILURE) {
				try {
					val result = state.lua.call(CoroutineLib.resume(), coroutine, childStatus.asBoolean)
					val execStatus = result[0] as Boolean

					if (!execStatus && result.size >= 2) {
						LOGGER.warn("Behavior DecoratorNode '$name' failed: ${result[1]}")
					}

					if (result.size == 1) {
						// another yield OR unexpected return?
						val coroutineStatus = state.lua.call(CoroutineLib.status(), coroutine)[0] as String

						if (coroutineStatus == "dead") {
							this.coroutine = null
							status = Status.SUCCESS
						} else
							status = Status.RUNNING
					} else {
						// yield or return with status
						val nodeStatus = result.getOrNull(1) as? Boolean ?: false
						// val nodeExtra = result.getOrNull(3) as? Table

						status = if (nodeStatus) Status.SUCCESS else Status.FAILURE
						this.coroutine = null
					}
				} catch (err: CallPausedException) {
					LOGGER.error("Behavior DecoratorNode '$name' called blocking code on children return, which initiated pause. This is not supported.")
					status = Status.FAILURE
				}
			} else {
				return Status.RUNNING
			}
		}

		return status
	}

	override fun reset() {
		coroutine = null
	}

	override fun toString(): String {
		return "DecoratorNode[$name, coroutine=$coroutine, parameters=$parameters, children=$child]"
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
